# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .log import Code, info, debug, debug_line, ModuleError
from . import Config, Result
from .loader import load
from .stages import main
from .saver import save
from .preview_creator import create_preview
from .utils import get_temp_folder
from .checker import check, check_equality
from .dsp import channel_count, size


def process(
        target: str,
        reference: str,
        results: list,
        config: Config = Config(),
        preview_target: Result = None,
        preview_result: Result = None
):
    debug('Please give us a star to help the project: https://github.com/sergree/matchering')
    debug_line()
    info(Code.INFO_LOADING)

    if not results:
        raise RuntimeError(f'The result list is empty')

    # Get a temporary folder for converting mp3's
    temp_folder = config.temp_folder if config.temp_folder else get_temp_folder(results)

    # Load the target
    target, target_sample_rate = load(target, 'target', temp_folder)
    # Analyze the target
    target, target_sample_rate = check(target, target_sample_rate, config, 'target')

    # Load the reference
    reference, reference_sample_rate = load(reference, 'reference', temp_folder)
    # Analyze the reference
    reference, reference_sample_rate = check(reference, reference_sample_rate, config, 'reference')

    # Analyze the target and the reference together
    if not config.allow_equality:
        check_equality(target, reference)

    # Validation of the most important conditions
    if not (target_sample_rate == reference_sample_rate == config.internal_sample_rate)\
            or not (channel_count(target) == channel_count(reference) == 2)\
            or not (size(target) > config.fft_size and size(reference) > config.fft_size):
        raise ModuleError(Code.ERROR_VALIDATION)

    # Process
    result, result_no_limiter, result_no_limiter_normalized = main(
        target,
        reference,
        config,
        need_default=any(rr.use_limiter for rr in results),
        need_no_limiter=any(not rr.use_limiter and not rr.normalize for rr in results),
        need_no_limiter_normalized=any(not rr.use_limiter and rr.normalize for rr in results),
    )

    debug_line()
    info(Code.INFO_EXPORTING)

    # Save
    for required_result in results:
        if required_result.use_limiter:
            correct_result = result
        else:
            if required_result.normalize:
                correct_result = result_no_limiter_normalized
            else:
                correct_result = result_no_limiter
        save(required_result.file, correct_result, config.internal_sample_rate, required_result.subtype)

    # Creating a preview (if needed)
    if preview_target or preview_result:
        result = next(item for item in [result, result_no_limiter, result_no_limiter_normalized] if item is not None)
        create_preview(target, result, config, preview_target, preview_result)

    debug_line()
    info(Code.INFO_COMPLETED)



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .explanations import get_explanation_handler


class __LogHandlers:
    @staticmethod
    def __dummy(*args, **kwargs):
        pass

    warning_handler = __dummy
    info_handler = __dummy
    debug_handler = __dummy
    explanation_handler = __dummy

    @staticmethod
    def __check_empty(explicit_handler, default_handler):
        return explicit_handler if explicit_handler else default_handler

    @classmethod
    def set_handlers(
            cls,
            default_handler=None,
            warning_handler=None,
            info_handler=None,
            debug_handler=None,
            show_codes=False
    ):
        default_handler = cls.__check_empty(default_handler, cls.__dummy)
        cls.warning_handler = cls.__check_empty(warning_handler, default_handler)
        cls.info_handler = cls.__check_empty(info_handler, default_handler)
        cls.debug_handler = cls.__check_empty(debug_handler, default_handler)
        cls.explanation_handler = get_explanation_handler(show_codes=show_codes)


def set_handlers(
        default_handler=None,
        warning_handler=None,
        info_handler=None,
        debug_handler=None,
        show_codes=False
):
    __LogHandlers.set_handlers(
        default_handler=default_handler,
        warning_handler=warning_handler,
        info_handler=info_handler,
        debug_handler=debug_handler,
        show_codes=show_codes
    )


def warning(*args, **kwargs):
    __LogHandlers.warning_handler(__LogHandlers.explanation_handler(*args, **kwargs))


def info(*args, **kwargs):
    __LogHandlers.info_handler(__LogHandlers.explanation_handler(*args, **kwargs))


def debug(*args, **kwargs):
    __LogHandlers.debug_handler(*args, **kwargs)


def debug_line():
    debug('-' * 40)



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import math
from scipy import signal
from scipy.ndimage.filters import maximum_filter1d

from .. import Config
from ..log import debug
from ..dsp import rectify, flip, max_mix
from ..utils import make_odd, ms_to_samples


def __sliding_window_fast(
        array: np.ndarray,
        window_size: int,
        mode: str = 'attack'
) -> np.ndarray:
    if mode == 'attack':
        window_size = make_odd(window_size)
        return maximum_filter1d(array, size=(2 * window_size - 1))
    half_window_size = (window_size - 1) // 2
    array = np.pad(array, (half_window_size, 0))
    return maximum_filter1d(array, size=window_size)[:-half_window_size]


def __process_attack(
        array: np.ndarray,
        config: Config
) -> (np.ndarray, np.ndarray):
    attack = ms_to_samples(config.limiter.attack, config.internal_sample_rate)

    slided_input = __sliding_window_fast(array, attack, mode='attack')

    coef = math.exp(config.limiter.attack_filter_coefficient / attack)
    b = [1 - coef]
    a = [1, -coef]
    output = signal.filtfilt(b, a, slided_input)

    return output, slided_input


def __process_release(
        array: np.ndarray,
        config: Config
) -> np.ndarray:
    hold = ms_to_samples(config.limiter.hold, config.internal_sample_rate)

    slided_input = __sliding_window_fast(array, hold, mode='hold')

    b, a = signal.butter(
        config.limiter.hold_filter_order,
        config.limiter.hold_filter_coefficient,
        fs=config.internal_sample_rate
    )
    hold_output = signal.lfilter(b, a, slided_input)

    b, a = signal.butter(
        config.limiter.release_filter_order,
        config.limiter.release_filter_coefficient / config.limiter.release,
        fs=config.internal_sample_rate
    )
    release_output = signal.lfilter(b, a, np.maximum(slided_input, hold_output))

    return np.maximum(hold_output, release_output)


def limit(
        array: np.ndarray,
        config: Config
) -> np.ndarray:

    debug('The limiter is started. Preparing the gain envelope...')
    rectified = rectify(array, config.threshold)

    if np.all(np.isclose(rectified, 1.)):
        debug('The limiter is not needed!')
        return array

    gain_hard_clip = flip(1. / rectified)
    debug('Modifying the gain envelope: attack stage...')
    gain_attack, gain_hard_clip_slided = __process_attack(np.copy(gain_hard_clip), config)

    debug('Modifying the gain envelope: hold / release stage...')
    gain_release = __process_release(np.copy(gain_hard_clip_slided), config)

    debug('Finalizing the gain envelope...')
    gain = flip(max_mix(gain_hard_clip, gain_attack, gain_release))

    return array * gain[:, None]




# -*- coding: utf-8 -*-

"""
Matchering Log
~~~~~~~~~~~~~~~~~~~~~

Internal Matchering logging module that can be bound to any text stream.

:copyright: (C) 2016-2020 Sergree
:license: GPLv3, see LICENSE for more details.

"""

from .codes import Code
from .handlers import warning, info, debug, debug_line
from .exceptions import ModuleError



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import random
import string
import math
from datetime import timedelta


def get_temp_folder(results: list) -> str:
    first_result_file = results[0].file
    return os.path.dirname(os.path.abspath(first_result_file))


def random_str(size: int = 16) -> str:
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=size))


def random_file(prefix: str = '', extension: str = 'wav') -> str:
    prefix = f'{prefix}-' if prefix else prefix
    return f'{prefix}{random_str()}.{extension}'


def __to_db_int(value: float) -> float:
    return 20 * math.log10(value)


def to_db(value: float) -> str:
    return f'{__to_db_int(value):.4f} dB'


def ms_to_samples(value: float, sample_rate: int) -> int:
    return int(sample_rate * value * 1e-3)


def make_odd(value: int) -> int:
    return value + 1 if not value & 1 else value


def time_str(length, sample_rate) -> str:
    return str(timedelta(seconds=length // sample_rate))



# -*- coding: utf-8 -*-

"""
Matchering Stages
~~~~~~~~~~~~~~~~~~~~~

Internal Matchering processing stage helpers.

:copyright: (C) 2016-2020 Sergree
:license: GPLv3, see LICENSE for more details.

"""

from .match_levels import normalize_reference, analyze_levels, get_average_rms, \
    get_lpis_and_match_rms, get_rms_c_and_amplify_pair
from .match_frequencies import get_fir, convolve



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from time import time
from scipy import signal, interpolate

from ..log import debug
from .. import Config
from ..dsp import ms_to_lr, smooth_lowess


def __average_fft(
        loudest_pieces: np.ndarray,
        sample_rate: int,
        fft_size: int
) -> np.ndarray:
    *_, specs = signal.stft(
        loudest_pieces,
        sample_rate,
        window='boxcar',
        nperseg=fft_size,
        noverlap=0,
        boundary=None,
        padded=False
    )
    return np.abs(specs).mean((0, 2))


def __smooth_exponentially(
        matching_fft: np.ndarray,
        config: Config
) -> np.ndarray:
    grid_linear = config.internal_sample_rate * 0.5 * np.linspace(
        0,
        1,
        config.fft_size // 2 + 1
    )

    grid_logarithmic = config.internal_sample_rate * 0.5 * np.logspace(
        np.log10(4 / config.fft_size),
        0,
        (config.fft_size // 2) * config.lin_log_oversampling + 1
    )

    interpolator = interpolate.interp1d(grid_linear, matching_fft, 'cubic')
    matching_fft_log = interpolator(grid_logarithmic)

    matching_fft_log_filtered = smooth_lowess(
        matching_fft_log,
        config.lowess_frac,
        config.lowess_it,
        config.lowess_delta
    )

    interpolator = interpolate.interp1d(grid_logarithmic, matching_fft_log_filtered, 'cubic', fill_value='extrapolate')
    matching_fft_filtered = interpolator(grid_linear)

    matching_fft_filtered[0] = 0
    matching_fft_filtered[1] = matching_fft[1]

    return matching_fft_filtered


def get_fir(
        target_loudest_pieces: np.ndarray,
        reference_loudest_pieces: np.ndarray,
        name: str,
        config: Config
) -> np.ndarray:
    debug(f'Calculating the {name} FIR for the matching EQ...')

    target_average_fft = __average_fft(target_loudest_pieces, config.internal_sample_rate, config.fft_size)
    reference_average_fft = __average_fft(reference_loudest_pieces, config.internal_sample_rate, config.fft_size)

    np.maximum(config.min_value, target_average_fft, out=target_average_fft)
    matching_fft = reference_average_fft / target_average_fft

    matching_fft_filtered = __smooth_exponentially(matching_fft, config)

    fir = np.fft.irfft(matching_fft_filtered)
    fir = np.fft.ifftshift(fir) * signal.windows.hann(len(fir))

    return fir


def convolve(
        target_mid: np.ndarray,
        mid_fir: np.ndarray,
        target_side: np.ndarray,
        side_fir: np.ndarray
) -> (np.ndarray, np.ndarray):
    debug('Convolving the TARGET audio with calculated FIRs...')
    timer = time()
    result_mid = signal.fftconvolve(target_mid, mid_fir, 'same')
    result_side = signal.fftconvolve(target_side, side_fir, 'same')
    debug(f'The convolution is done in {time() - timer:.2f} seconds')

    debug('Converting MS to LR...')
    result = ms_to_lr(result_mid, result_side)

    return result, result_mid



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import statsmodels.api as sm


def size(array: np.ndarray) -> int:
    return array.shape[0]


def channel_count(array: np.ndarray) -> int:
    return array.shape[1]


def is_mono(array: np.ndarray) -> bool:
    return array.shape[1] == 1


def is_stereo(array: np.ndarray) -> bool:
    return array.shape[1] == 2


def is_1d(array: np.ndarray) -> bool:
    return len(array.shape) == 1


def mono_to_stereo(array: np.ndarray) -> np.ndarray:
    return np.repeat(array, repeats=2, axis=1)


def count_max_peaks(array: np.ndarray) -> (float, int):
    max_value = np.abs(array).max()
    max_count = np.count_nonzero(
        np.logical_or(
            np.isclose(array, max_value),
            np.isclose(array, -max_value)
        )
    )
    return max_value, max_count


def lr_to_ms(array: np.ndarray) -> (np.ndarray, np.ndarray):
    array = np.copy(array)
    array[:, 0] += array[:, 1]
    array[:, 0] *= 0.5
    mid = np.copy(array[:, 0])
    array[:, 0] -= array[:, 1]
    side = np.copy(array[:, 0])
    return mid, side


def ms_to_lr(mid_array: np.ndarray, side_array: np.ndarray) -> np.ndarray:
    return np.vstack((mid_array + side_array, mid_array - side_array)).T


def unfold(array: np.ndarray, piece_size: int, divisions: int) -> np.ndarray:
    # (len(array),) -> (divisions, piece_size)
    return array[:piece_size * divisions].reshape(-1, piece_size)


def rms(array: np.ndarray) -> float:
    return np.sqrt(array @ array / array.shape[0])


def batch_rms(array: np.ndarray) -> np.ndarray:
    piece_size = array.shape[1]
    # (divisions, piece_size) -> (divisions, 1, piece_size)
    multiplicand = array[:, None, :]
    # (divisions, piece_size) -> (divisions, piece_size, 1)
    multiplier = array[..., None]
    return np.sqrt(np.squeeze(multiplicand @ multiplier, axis=(1, 2)) / piece_size)


def amplify(array: np.ndarray, gain: float) -> np.ndarray:
    return array * gain


def normalize(
        array: np.ndarray,
        threshold: float,
        epsilon: float,
        normalize_clipped: bool
):
    coefficient = 1.
    max_value = np.abs(array).max()
    if max_value < threshold or normalize_clipped:
        coefficient = max(epsilon, max_value / threshold)
    return array / coefficient, coefficient


def smooth_lowess(
        array: np.ndarray,
        frac: float,
        it: int,
        delta: float
) -> np.ndarray:
    return sm.nonparametric.lowess(
        array,
        np.linspace(0, 1, len(array)),
        frac=frac,
        it=it,
        delta=delta
    )[:, 1]


def clip(array: np.ndarray) -> np.ndarray:
    return np.clip(array, -1, 1)


def flip(array: np.ndarray) -> np.ndarray:
    return 1. - array


def rectify(
        array: np.ndarray,
        threshold: float
) -> np.ndarray:
    rectified = np.abs(array).max(1)
    rectified[rectified <= threshold] = threshold
    rectified /= threshold
    return rectified


def max_mix(*args) -> np.ndarray:
    return np.maximum.reduce(args)


def strided_app_2d(matrix: np.ndarray, batch_size: int, step: int) -> np.ndarray:
    matrix_length = matrix.shape[0]
    matrix_width = matrix.shape[1]
    if batch_size > matrix_length:
        return np.expand_dims(matrix, axis=0)
    batch_count = ((matrix_length - batch_size) // step) + 1
    stride_length, stride_width = matrix.strides
    return np.lib.stride_tricks.as_strided(
        matrix,
        shape=(batch_count, batch_size, matrix_width),
        strides=(step * stride_length, stride_length, stride_width)
    )


def batch_rms_2d(array: np.ndarray) -> np.ndarray:
    return batch_rms(array.reshape(array.shape[0], array.shape[1] * array.shape[2]))


def fade(array: np.ndarray, fade_size: int) -> np.ndarray:
    array = np.copy(array)
    fade_in = np.linspace(0, 1, fade_size)
    fade_out = fade_in[::-1]
    array[:fade_size].T[:] *= fade_in
    array[size(array) - fade_size:].T[:] *= fade_out
    return array



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from resampy import resample

from .log import Code, warning, info, debug, ModuleError
from . import Config
from .dsp import size, is_mono, is_stereo, mono_to_stereo, count_max_peaks
from .utils import time_str


def __check_sample_rate(
        array: np.ndarray,
        sample_rate: int,
        required_sample_rate: int,
        name: str,
        log_handler,
        log_code: Code
) -> (np.ndarray, int):
    if sample_rate != required_sample_rate:
        debug(f'Resampling {name} audio from {sample_rate} Hz to {required_sample_rate} Hz...')
        array = resample(array, sample_rate, required_sample_rate, axis=0)
        log_handler(log_code)
    return array, required_sample_rate


def __check_length(
        array: np.ndarray,
        sample_rate: int,
        max_length: int,
        min_length: int,
        name: str,
        error_code_max: Code,
        error_code_min: Code
) -> None:
    length = size(array)
    debug(f'{name} audio length: {length} samples ({time_str(length, sample_rate)})')
    if length > max_length:
        raise ModuleError(error_code_max)
    elif length < min_length:
        raise ModuleError(error_code_min)


def __check_channels(
        array: np.ndarray,
        info_code_mono: Code,
        error_code_not_stereo: Code
) -> np.ndarray:
    if is_mono(array):
        info(info_code_mono)
        array = mono_to_stereo(array)
    elif not is_stereo(array):
        raise ModuleError(error_code_not_stereo)
    return array


def __check_clipping_limiting(
        array: np.ndarray,
        clipping_samples_threshold: int,
        limited_samples_threshold: int,
        warning_code_clipping: Code,
        warning_code_limiting: Code
) -> None:
    max_value, max_count = count_max_peaks(array)
    if max_count > clipping_samples_threshold:
        if np.isclose(max_value, 1.):
            warning(warning_code_clipping)
        elif max_count > limited_samples_threshold:
            warning(warning_code_limiting)


def check(array: np.ndarray, sample_rate: int, config: Config, name: str) -> (np.ndarray, int):
    name = name.upper()

    __check_length(
        array,
        sample_rate,
        config.max_length * sample_rate,
        config.fft_size * sample_rate // config.internal_sample_rate,
        name,
        Code.ERROR_TARGET_LENGTH_IS_EXCEEDED if name == 'TARGET'
        else Code.ERROR_REFERENCE_LENGTH_LENGTH_IS_EXCEEDED,
        Code.ERROR_TARGET_LENGTH_IS_TOO_SMALL if name == 'TARGET'
        else Code.ERROR_REFERENCE_LENGTH_LENGTH_TOO_SMALL
    )

    array = __check_channels(
        array,
        Code.INFO_TARGET_IS_MONO if name == 'TARGET' else Code.INFO_REFERENCE_IS_MONO,
        Code.ERROR_TARGET_NUM_OF_CHANNELS_IS_EXCEEDED if name == 'TARGET'
        else Code.ERROR_REFERENCE_NUM_OF_CHANNELS_IS_EXCEEDED
    )

    array, sample_rate = __check_sample_rate(
        array,
        sample_rate,
        config.internal_sample_rate,
        name,
        warning if name == 'TARGET' else info,
        Code.WARNING_TARGET_IS_RESAMPLED if name == 'TARGET'
        else Code.INFO_REFERENCE_IS_RESAMPLED
    )

    if name == 'TARGET':
        __check_clipping_limiting(
            array,
            config.clipping_samples_threshold,
            config.limited_samples_threshold,
            Code.WARNING_TARGET_IS_CLIPPING,
            Code.WARNING_TARGET_LIMITER_IS_APPLIED
        )

    return array, sample_rate


def check_equality(target: np.ndarray, reference: np.ndarray) -> None:
    if target.shape == reference.shape and np.allclose(target, reference):
        raise ModuleError(Code.ERROR_TARGET_EQUALS_REFERENCE)



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import soundfile as sf

from .log import debug


def save(
        file: str,
        result: np.ndarray,
        sample_rate: int,
        subtype: str,
        name: str = 'result'
) -> None:
    name = name.upper()
    debug(f'Saving the {name} {sample_rate} Hz Stereo {subtype} to: \'{file}\'...')
    sf.write(file, result, sample_rate, subtype)
    debug(f'\'{file}\' is saved')



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .explanations import get_explanation_handler
from .codes import Code


class ModuleError(Exception):
    def __init__(self, code: Code):
        Exception.__init__(self, get_explanation_handler(show_codes=True)(code))



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy as np
import soundfile as sf
import subprocess

from .log import Code, warning, info, debug, ModuleError
from .utils import random_file


def load(file: str, file_type: str, temp_folder: str) -> (np.ndarray, int):
    file_type = file_type.upper()
    sound, sample_rate = None, None
    debug(f'Loading the {file_type} file: \'{file}\'...')
    try:
        sound, sample_rate = sf.read(file, always_2d=True)
    except RuntimeError as e:
        debug(e)
        if 'unknown format' in str(e):
            sound, sample_rate = __load_with_ffmpeg(file, file_type, temp_folder)
    if sound is None or sample_rate is None:
        if file_type == 'TARGET':
            raise ModuleError(Code.ERROR_TARGET_LOADING)
        else:
            raise ModuleError(Code.ERROR_REFERENCE_LOADING)
    debug(f'The {file_type} file is loaded')
    return sound, sample_rate


def __load_with_ffmpeg(file: str, file_type: str, temp_folder: str) -> (np.ndarray, int):
    sound, sample_rate = None, None
    debug(f'Trying to load \'{file}\' with ffmpeg...')
    temp_file = os.path.join(temp_folder, random_file(prefix='temp'))
    with open(os.devnull, 'w') as devnull:
        try:
            subprocess.check_call(
                [
                    'ffmpeg',
                    '-i',
                    file,
                    temp_file
                ],
                stdout=devnull,
                stderr=devnull
            )
            sound, sample_rate = sf.read(temp_file, always_2d=True)
            if file_type == 'TARGET':
                warning(Code.WARNING_TARGET_IS_LOSSY)
            else:
                info(Code.INFO_REFERENCE_IS_LOSSY)
            os.remove(temp_file)
        except FileNotFoundError:
            debug('ffmpeg is not found in the system! '
                  'Download, install and add it to PATH: https://www.ffmpeg.org/download.html')
        except subprocess.CalledProcessError:
            debug(f'ffmpeg cannot convert \'{file}\' to .wav!')
    return sound, sample_rate



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from .log import Code, info, debug, debug_line
from . import Config
from .utils import to_db
from .dsp import amplify, normalize, clip
from .stage_helpers import normalize_reference, analyze_levels, get_fir, \
    convolve, get_average_rms, get_lpis_and_match_rms, get_rms_c_and_amplify_pair
from .limiter import limit


def __match_levels(
        target: np.ndarray,
        reference: np.ndarray,
        config: Config
) -> (np.ndarray, np.ndarray, float, np.ndarray, np.ndarray, np.ndarray, np.ndarray, float, float):
    debug_line()
    info(Code.INFO_MATCHING_LEVELS)

    debug(f'The maximum size of the analyzed piece: {config.max_piece_size} samples '
          f'or {config.max_piece_size / config.internal_sample_rate:.2f} seconds')

    reference, final_amplitude_coefficient = normalize_reference(reference, config)

    target_mid, target_side,\
        target_mid_loudest_pieces, target_side_loudest_pieces,\
        target_match_rms, target_divisions, target_piece_size\
        = analyze_levels(target, 'target', config)

    reference_mid, reference_side,\
        reference_mid_loudest_pieces, reference_side_loudest_pieces,\
        reference_match_rms, *_\
        = analyze_levels(reference, 'reference', config)

    rms_coefficient, target_mid, target_side = get_rms_c_and_amplify_pair(
        target_mid, target_side,
        target_match_rms, reference_match_rms,
        config.min_value, 'target'
    )

    debug('Modifying the amplitudes of the extracted loudest TARGET pieces...')
    target_mid_loudest_pieces = amplify(target_mid_loudest_pieces, rms_coefficient)
    target_side_loudest_pieces = amplify(target_side_loudest_pieces, rms_coefficient)

    return target_mid, target_side, final_amplitude_coefficient,\
        target_mid_loudest_pieces, target_side_loudest_pieces,\
        reference_mid_loudest_pieces, reference_side_loudest_pieces,\
        target_divisions, target_piece_size, reference_match_rms


def __match_frequencies(
        target_mid: np.ndarray,
        target_side: np.ndarray,
        target_mid_loudest_pieces: np.ndarray,
        reference_mid_loudest_pieces: np.ndarray,
        target_side_loudest_pieces: np.ndarray,
        reference_side_loudest_pieces: np.ndarray,
        config: Config
) -> (np.ndarray, np.ndarray):
    debug_line()
    info(Code.INFO_MATCHING_FREQS)

    mid_fir = get_fir(
        target_mid_loudest_pieces, reference_mid_loudest_pieces,
        'mid',
        config
    )
    side_fir = get_fir(
        target_side_loudest_pieces, reference_side_loudest_pieces,
        'side',
        config
    )

    result, result_mid = convolve(target_mid, mid_fir, target_side, side_fir)

    return result, result_mid


def __correct_levels(
        result: np.ndarray,
        result_mid: np.ndarray,
        target_divisions: int,
        target_piece_size: int,
        reference_match_rms: float,
        config: Config
) -> np.ndarray:
    debug_line()
    info(Code.INFO_CORRECTING_LEVELS)

    for step in range(1, config.rms_correction_steps + 1):
        debug(f'Applying RMS correction #{step}...')
        result_mid_clipped = clip(result_mid)

        _, clipped_rmses, clipped_average_rms = get_average_rms(
            result_mid_clipped,
            target_piece_size,
            target_divisions,
            'result'
        )

        _, result_mid_clipped_match_rms = get_lpis_and_match_rms(clipped_rmses, clipped_average_rms)

        rms_coefficient, result_mid, result = get_rms_c_and_amplify_pair(
            result_mid, result,
            result_mid_clipped_match_rms, reference_match_rms,
            config.min_value, 'result'
        )

    return result


def __finalize(
    result_no_limiter: np.ndarray,
    final_amplitude_coefficient: float,
    need_default: bool,
    need_no_limiter: bool,
    need_no_limiter_normalized: bool,
    config: Config
) -> (np.ndarray, np.ndarray, np.ndarray):
    debug_line()
    info(Code.INFO_FINALIZING)

    result_no_limiter_normalized = None
    if need_no_limiter_normalized:
        result_no_limiter_normalized, coefficient = normalize(
            result_no_limiter,
            config.threshold,
            config.min_value,
            normalize_clipped=True
        )
        debug(f'The amplitude of the normalized RESULT should be adjusted by {to_db(coefficient)}')
        if not np.isclose(final_amplitude_coefficient, 1.):
            debug(f'And by {to_db(final_amplitude_coefficient)} after applying some brickwall limiter to it')

    result = None
    if need_default:
        result = limit(result_no_limiter, config)
        result = amplify(result, final_amplitude_coefficient)

    result_no_limiter = result_no_limiter if need_no_limiter else None

    return result, result_no_limiter, result_no_limiter_normalized


def main(
        target: np.ndarray,
        reference: np.ndarray,
        config: Config,
        need_default: bool = True,
        need_no_limiter: bool = False,
        need_no_limiter_normalized: bool = False
) -> (np.ndarray, np.ndarray, np.ndarray):
    target_mid, target_side, final_amplitude_coefficient,\
        target_mid_loudest_pieces, target_side_loudest_pieces,\
        reference_mid_loudest_pieces, reference_side_loudest_pieces,\
        target_divisions, target_piece_size, reference_match_rms\
        = __match_levels(target, reference, config)

    del target, reference

    result_no_limiter, result_no_limiter_mid = __match_frequencies(
        target_mid, target_side,
        target_mid_loudest_pieces, reference_mid_loudest_pieces,
        target_side_loudest_pieces, reference_side_loudest_pieces,
        config
    )

    del target_mid, target_side,\
        target_mid_loudest_pieces, reference_mid_loudest_pieces,\
        target_side_loudest_pieces, reference_side_loudest_pieces

    result_no_limiter = __correct_levels(
        result_no_limiter, result_no_limiter_mid,
        target_divisions, target_piece_size,
        reference_match_rms,
        config
    )

    del result_no_limiter_mid

    result, result_no_limiter, result_no_limiter_normalized = __finalize(
        result_no_limiter, final_amplitude_coefficient,
        need_default, need_no_limiter, need_no_limiter_normalized,
        config
    )

    return result, result_no_limiter, result_no_limiter_normalized



# -*- coding: utf-8 -*-

"""
Matchering Hyrax Limiter
~~~~~~~~~~~~~~~~~~~~~

Internal Matchering brickwall limiter based on the use of time constant and Butterworth filters with sliding Windows.

:copyright: (C) 2016-2020 Sergree
:license: GPLv3, see LICENSE for more details.

"""

from .hyrax import limit



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np

from .log import Code, info, debug, debug_line
from .dsp import size, strided_app_2d, batch_rms_2d, fade
from . import Config, Result
from .saver import save
from .utils import time_str


def create_preview(
        target: np.ndarray,
        result: np.ndarray,
        config: Config,
        preview_target: Result,
        preview_result: Result
) -> None:
    debug_line()
    info(Code.INFO_MAKING_PREVIEWS)

    debug(f'The maximum duration of the preview is {config.preview_size / config.internal_sample_rate} seconds, '
          f'with the analysis step of {config.preview_analysis_step / config.internal_sample_rate} seconds')

    target_pieces = strided_app_2d(target, config.preview_size, config.preview_analysis_step)
    result_pieces = strided_app_2d(result, config.preview_size, config.preview_analysis_step)

    result_loudest_piece_idx = np.argmax(batch_rms_2d(result_pieces))

    target_piece = target_pieces[result_loudest_piece_idx]
    result_piece = result_pieces[result_loudest_piece_idx]

    debug_sample_begin = config.preview_analysis_step * int(result_loudest_piece_idx)
    debug_sample_end = debug_sample_begin + size(result_piece)
    debug(f'The best part to preview: '
          f'{time_str(debug_sample_begin, config.internal_sample_rate)} '
          f'- {time_str(debug_sample_end, config.internal_sample_rate)}')

    if size(result) != size(result_piece):
        fade_size = min(config.preview_fade_size, size(result_piece) // config.preview_fade_coefficient)
        target_piece, result_piece = fade(target_piece, fade_size), fade(result_piece, fade_size)

    if preview_target:
        save(preview_target.file, target_piece, config.internal_sample_rate, preview_target.subtype, 'target preview')

    if preview_result:
        save(preview_result.file, result_piece, config.internal_sample_rate, preview_result.subtype, 'result preview')



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np

from ..log import debug
from .. import Config
from ..utils import to_db
from ..dsp import lr_to_ms, size, unfold, batch_rms, rms, amplify, normalize


def normalize_reference(
        reference: np.ndarray,
        config: Config
) -> (np.ndarray, float):
    debug('Normalizing the REFERENCE...')

    reference, final_amplitude_coefficient = normalize(
        reference,
        config.threshold,
        config.min_value,
        normalize_clipped=False
    )

    if np.isclose(final_amplitude_coefficient, 1.):
        debug('The REFERENCE was not changed. There is no final amplitude coefficient')
    else:
        debug(f'The REFERENCE was normalized. '
              f'Final amplitude coefficient for the TARGET audio is: {to_db(final_amplitude_coefficient)}')

    return reference, final_amplitude_coefficient


def __calculate_piece_sizes(
        array: np.ndarray,
        max_piece_size: int,
        name: str,
        sample_rate: int
) -> (int, int, int):
    array_size = size(array)
    divisions = int(array_size / max_piece_size) + 1
    debug(f'The {name} will be didived into {divisions} pieces')

    piece_size = int(array_size / divisions)
    debug(f'One piece of the {name} has a length of {piece_size} samples or {piece_size / sample_rate:.2f} seconds')

    return array_size, divisions, piece_size


def get_lpis_and_match_rms(
        rmses: np.ndarray,
        average_rms: float
) -> (np.ndarray, float):
    loudest_piece_idxs = np.where(rmses >= average_rms)

    loudest_rmses = rmses[loudest_piece_idxs]
    match_rms = rms(loudest_rmses)
    debug(f'The current average RMS value in the loudest pieces is {to_db(match_rms)}')

    return loudest_piece_idxs, match_rms


def __extract_loudest_pieces(
        rmses: np.ndarray,
        average_rms: float,
        unfolded_mid: np.ndarray,
        unfolded_side: np.ndarray,
        name: str
) -> (np.ndarray, np.ndarray, float):
    debug(f'Extracting the loudest pieces of the {name} audio '
          f'with the RMS value more than average {to_db(average_rms)}...')
    loudest_piece_idxs, match_rms = get_lpis_and_match_rms(rmses, average_rms)

    mid_loudest_pieces = unfolded_mid[loudest_piece_idxs]
    side_loudest_pieces = unfolded_side[loudest_piece_idxs]

    return mid_loudest_pieces, side_loudest_pieces, match_rms


def get_average_rms(
        array: np.ndarray,
        piece_size: int,
        divisions: int,
        name: str
) -> (np.ndarray, np.ndarray, float):
    name = name.upper()
    unfolded_array = unfold(array, piece_size, divisions)

    debug(f'Calculating RMSes of the {name} pieces...')
    rmses = batch_rms(unfolded_array)
    average_rms = rms(rmses)

    return unfolded_array, rmses, average_rms


def __calculate_rms_coefficient(
        array_match_rms: float,
        reference_match_rms: float,
        epsilon: float
) -> float:
    rms_coefficient = reference_match_rms / max(epsilon, array_match_rms)
    debug(f'The RMS coefficient is: {to_db(rms_coefficient)}')
    return rms_coefficient


def get_rms_c_and_amplify_pair(
        array_main: np.ndarray,
        array_additional: np.ndarray,
        array_main_match_rms: float,
        reference_match_rms: float,
        epsilon: float,
        name: str
) -> (float, np.ndarray, np.ndarray):
    name = name.upper()
    rms_coefficient = __calculate_rms_coefficient(array_main_match_rms, reference_match_rms, epsilon)

    debug(f'Modifying the amplitudes of the {name} audio...')
    array_main = amplify(array_main, rms_coefficient)
    array_additional = amplify(array_additional, rms_coefficient)

    return rms_coefficient, array_main, array_additional


def analyze_levels(
        array: np.ndarray,
        name: str,
        config: Config
) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, float, float, float):
    name = name.upper()
    debug(f'Calculating mid and side channels of the {name}...')
    mid, side = lr_to_ms(array)
    del array

    array_size, divisions, piece_size = __calculate_piece_sizes(
        mid,
        config.max_piece_size,
        name,
        config.internal_sample_rate
    )

    unfolded_mid, rmses, average_rms = get_average_rms(mid, piece_size, divisions, name)
    unfolded_side = unfold(side, piece_size, divisions)

    mid_loudest_pieces, side_loudest_pieces, match_rms = __extract_loudest_pieces(
        rmses,
        average_rms,
        unfolded_mid,
        unfolded_side,
        name
    )

    return mid, side, mid_loudest_pieces, side_loudest_pieces, match_rms, divisions, piece_size



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='matchering',
    version='2.0.0',
    author='Sergree',
    author_email='wokashi.rg@gmail.com',
    description='Audio Matching and Mastering Python Library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    license='GPLv3',
    url='https://github.com/sergree/matchering',
    packages=find_packages(include=['matchering', 'matchering.*']),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Natural Language :: English',
        'Topic :: Artistic Software',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Multimedia :: Sound/Audio :: Conversion',
        'Topic :: Utilities'
    ],
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from enum import IntEnum


class Code(IntEnum):
    INFO_UPLOADING = 2001
    INFO_WAITING = 2002
    INFO_LOADING = 2003
    INFO_MATCHING_LEVELS = 2004
    INFO_MATCHING_FREQS = 2005
    INFO_CORRECTING_LEVELS = 2006
    INFO_FINALIZING = 2007
    INFO_EXPORTING = 2008
    INFO_MAKING_PREVIEWS = 2009
    INFO_COMPLETED = 2010

    INFO_TARGET_IS_MONO = 2101
    INFO_REFERENCE_IS_MONO = 2201
    INFO_REFERENCE_IS_RESAMPLED = 2202
    INFO_REFERENCE_IS_LOSSY = 2203

    WARNING_TARGET_IS_CLIPPING = 3001
    WARNING_TARGET_LIMITER_IS_APPLIED = 3002
    WARNING_TARGET_IS_RESAMPLED = 3003
    WARNING_TARGET_IS_LOSSY = 3004

    ERROR_TARGET_LOADING = 4001
    ERROR_TARGET_LENGTH_IS_EXCEEDED = 4002
    ERROR_TARGET_LENGTH_IS_TOO_SMALL = 4003
    ERROR_TARGET_NUM_OF_CHANNELS_IS_EXCEEDED = 4004
    ERROR_TARGET_EQUALS_REFERENCE = 4005

    ERROR_REFERENCE_LOADING = 4101
    ERROR_REFERENCE_LENGTH_LENGTH_IS_EXCEEDED = 4102
    ERROR_REFERENCE_LENGTH_LENGTH_TOO_SMALL = 4103
    ERROR_REFERENCE_NUM_OF_CHANNELS_IS_EXCEEDED = 4104

    ERROR_UNKNOWN = 4201
    ERROR_VALIDATION = 4202



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .codes import Code


def __default(code: Code) -> str:
    return __en[code]


def __verbose(code: Code) -> str:
    return f'{code}: {__en[code]}'


def get_explanation_handler(show_codes: bool = False):
    return __verbose if show_codes else __default


__en = {
    Code.INFO_UPLOADING: 'Uploading files',
    Code.INFO_WAITING: 'Queued for processing',
    Code.INFO_LOADING: 'Loading and analysis',
    Code.INFO_MATCHING_LEVELS: 'Matching levels',
    Code.INFO_MATCHING_FREQS: 'Matching frequencies',
    Code.INFO_CORRECTING_LEVELS: 'Correcting levels',
    Code.INFO_FINALIZING: 'Final processing and saving',
    Code.INFO_EXPORTING: 'Exporting various audio formats',
    Code.INFO_MAKING_PREVIEWS: 'Making previews',
    Code.INFO_COMPLETED: 'The task is completed',

    Code.INFO_TARGET_IS_MONO: 'The TARGET audio is mono. Converting it to stereo...',
    Code.INFO_REFERENCE_IS_MONO: 'The REFERENCE audio is mono. Converting it to stereo...',
    Code.INFO_REFERENCE_IS_RESAMPLED: 'The REFERENCE audio was resampled',
    Code.INFO_REFERENCE_IS_LOSSY: 'Presumably the REFERENCE audio format is lossy',

    Code.WARNING_TARGET_IS_CLIPPING: 'Audio clipping is detected in the TARGET file. '
                                     'It is highly recommended to use the non-clipping version',
    Code.WARNING_TARGET_LIMITER_IS_APPLIED: 'The applied limiter is detected in the TARGET file. '
                                            'It is highly recommended to use the version without a limiter',
    Code.WARNING_TARGET_IS_RESAMPLED: 'The TARGET audio sample rate and internal sample rate were different. '
                                      'The TARGET audio was resampled',
    Code.WARNING_TARGET_IS_LOSSY: 'Presumably the TARGET audio format is lossy. '
                                  'It is highly recommended to use lossless audio formats (WAV, FLAC, AIFF)',

    Code.ERROR_TARGET_LOADING: 'Audio stream error in the TARGET file',
    Code.ERROR_TARGET_LENGTH_IS_EXCEEDED: 'Track length is exceeded in the TARGET file',
    Code.ERROR_TARGET_LENGTH_IS_TOO_SMALL: 'The track length is too small in the TARGET file',
    Code.ERROR_TARGET_NUM_OF_CHANNELS_IS_EXCEEDED: 'The number of channels exceeded in the TARGET file',
    Code.ERROR_TARGET_EQUALS_REFERENCE: 'The TARGET and REFERENCE files are the same. '
                                        'They must be different so that Matchering makes sense',

    Code.ERROR_REFERENCE_LOADING: 'Audio stream error in the REFERENCE file',
    Code.ERROR_REFERENCE_LENGTH_LENGTH_IS_EXCEEDED: 'Track length is exceeded in the REFERENCE file',
    Code.ERROR_REFERENCE_LENGTH_LENGTH_TOO_SMALL: 'The track length is too small in the REFERENCE file',
    Code.ERROR_REFERENCE_NUM_OF_CHANNELS_IS_EXCEEDED: 'The number of channels exceeded in the REFERENCE file',

    Code.ERROR_UNKNOWN: 'Unknown error',
    Code.ERROR_VALIDATION: 'Validation failed! Please let the developers know about this error!'
}



# -*- coding: utf-8 -*-

"""
Matchering - Audio Matching and Mastering Python Library
Copyright (C) 2016-2020 Sergree

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import math
from .log import debug


class LimiterConfig:
    def __init__(
            self,
            attack: float = 1,
            hold: float = 1,
            release: float = 3000,
            attack_filter_coefficient: float = -2,
            hold_filter_order: int = 1,
            hold_filter_coefficient: float = 7,
            release_filter_order: int = 1,
            release_filter_coefficient: float = 800
    ):
        assert attack > 0
        self.attack = attack

        assert hold > 0
        self.hold = hold

        assert release > 0
        self.release = release

        self.attack_filter_coefficient = attack_filter_coefficient

        assert hold_filter_order > 0
        assert isinstance(hold_filter_order, int)
        self.hold_filter_order = hold_filter_order

        self.hold_filter_coefficient = hold_filter_coefficient

        assert release_filter_order > 0
        assert isinstance(release_filter_order, int)
        self.release_filter_order = release_filter_order

        self.release_filter_coefficient = release_filter_coefficient


class Config:
    def __init__(
            self,
            internal_sample_rate: int = 44100,
            max_length: float = 15 * 60,
            max_piece_size: float = 15,
            threshold: float = (2 ** 15 - 61) / 2 ** 15,
            min_value: float = 1e-6,
            fft_size: int = 4096,
            lin_log_oversampling: int = 4,
            rms_correction_steps: int = 4,
            clipping_samples_threshold: int = 8,
            limited_samples_threshold: int = 128,
            allow_equality: bool = False,
            lowess_frac: float = 0.0375,
            lowess_it: int = 0,
            lowess_delta: float = 0.001,
            preview_size: float = 30,
            preview_analysis_step: float = 5,
            preview_fade_size: float = 1,
            preview_fade_coefficient: float = 8,
            temp_folder: str = None,
            limiter: LimiterConfig = LimiterConfig(),
    ):
        assert internal_sample_rate > 0
        assert isinstance(internal_sample_rate, int)
        if internal_sample_rate != 44100:
            debug('Using an internal sample rate other than 44100 has not been tested properly! '
                  'Use it at your own risk!')
        self.internal_sample_rate = internal_sample_rate

        assert max_length > 0
        assert max_length > fft_size / internal_sample_rate
        self.max_length = max_length

        assert threshold > min_value
        assert threshold < 1
        self.threshold = threshold

        assert min_value > 0
        assert min_value < 0.1
        self.min_value = min_value

        assert max_piece_size > 0
        assert max_piece_size > fft_size / internal_sample_rate
        assert max_piece_size < max_length
        self.max_piece_size = max_piece_size * internal_sample_rate

        assert fft_size > 1
        assert math.log2(fft_size).is_integer()
        self.fft_size = fft_size

        assert lin_log_oversampling > 0
        assert isinstance(lin_log_oversampling, int)
        self.lin_log_oversampling = lin_log_oversampling

        assert rms_correction_steps >= 0
        assert isinstance(rms_correction_steps, int)
        self.rms_correction_steps = rms_correction_steps

        assert clipping_samples_threshold >= 0
        assert limited_samples_threshold > 0
        assert limited_samples_threshold > clipping_samples_threshold
        assert isinstance(clipping_samples_threshold, int)
        assert isinstance(limited_samples_threshold, int)
        self.clipping_samples_threshold = clipping_samples_threshold
        self.limited_samples_threshold = limited_samples_threshold

        assert isinstance(allow_equality, bool)
        self.allow_equality = allow_equality

        assert lowess_frac > 0
        assert lowess_it >= 0
        assert lowess_delta >= 0
        assert isinstance(lowess_it, int)
        self.lowess_frac = lowess_frac
        self.lowess_it = lowess_it
        self.lowess_delta = lowess_delta

        assert preview_size > 5
        assert preview_analysis_step > 1
        assert preview_fade_size > 0
        assert preview_fade_coefficient >= 2
        self.preview_size = preview_size * internal_sample_rate
        self.preview_analysis_step = preview_analysis_step * internal_sample_rate
        self.preview_fade_size = preview_fade_size * internal_sample_rate
        self.preview_fade_coefficient = preview_fade_coefficient

        assert temp_folder is None or isinstance(temp_folder, str)
        self.temp_folder = temp_folder

        assert isinstance(limiter, LimiterConfig)
        self.limiter = limiter



# -*- coding: utf-8 -*-

"""
Matchering
~~~~~~~~~~~~~~~~~~~

Audio Matching and Mastering Python Library

:copyright: (C) 2016-2020 Sergree
:license: GPLv3, see LICENSE for more details.

"""

__title__ = 'matchering'

__author__ = 'Sergree'
__credits__ = ['Sergey Grishakov', 'Igor Isaev', 'Chin Yun Yu']
__maintainer__ = 'Sergree'
__email__ = 'wokashi.rg@gmail.com'
__license__ = 'GPLv3'
__copyright__ = 'Copyright (C) 2016-2020 Sergree'

__version__ = '2.0.0'

from .log.handlers import set_handlers as log
from .results import Result, pcm16, pcm24
from .defaults import Config
from .core import process
from .loader import load
from .checker import check
